#!/usr/local/bin/apl --script
⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝
⍝
⍝ tb 2017-06-27 14:56:12 (GMT-4)
⍝
⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝

)copy 1 import
)copy 1 wp
)copy 1 prompt

⍝ tb

⍝ This workspaceprovides functions to load and adjust trial balances.
⍝ A trial balance is a list of accounts together with their balances
⍝ to demonstrate that a company's books of account balances.  That the
⍝ debits equal the credits.  Accountants use trial balances to prepare
⍝ financial reports.

⍝ Doc is the basic unit of data for the trial balance.  It is a two
⍝ element vector made up of the head and body.  The Head it self is a
⍝ vector of doc_id, Journal, Name, Date and Description.  Once posted
⍝ the head becomes a line in the db_Documents table.

⍝ The body is a list of lines.  Each line is a vector of five items:
⍝ doc_id (to link it to the tb_Documents table), line_no, Acct_no (a
⍝ pointer to the tb_Accounts table), Debit and Credit.  Once posted
⍝ each line is appended to the tb_DocLines table.

⍝ Each document is uniquely identified by it's journal, name and date.

⍝ tb an application to demonstrate use
⍝ Copyright (C) 2018 Bill Daly

⍝ This program is free software: you can redistribute it and/or modify
⍝ it under the terms of the GNU General Public License as published by
⍝ the Free Software Foundation, either version 3 of the License, or
⍝ (at your option) any later version.

⍝ This program is distributed in the hope that it will be useful,
⍝ but WITHOUT ANY WARRANTY; without even the implied warranty of
⍝ MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
⍝ GNU General Public License for more details.

⍝ You should have received a copy of the GNU General Public License
⍝ along with this program.  If not, see <http://www.gnu.org/licenses/>.

∇tb_acct_add acct;re;locale
  ⍝ Function adds an account to tb_accounts. At a minimum aact must be
  ⍝ a two element vector of account number and title. The account
  ⍝ type, element 3, and sign type, element 4 may also be supplied.
  re←(tb_accounts[;3]='r')/tb_accounts[;1]
  ⎕es (~util∆numberp acct[1])/'THE ACCOUNT NUMBER MUST BE A NUMBER'
  ⎕es (' '≠1↑0⍴⊃acct[2])/'THE ACCOUNT TITLE MUST BE TEXT'
  ⎕es (acct[1]∊tb_accounts[;1])/(⍕acct[1]),' IS CURRENTLY IN THE CHART OF ACCOUNTS'
  →(4=⍴acct←,acct)/just_add
  →(3=⍴acct)/with_acct_type
  ⍎(re>acct[1])/'acct←acct,''b''◊→with_acct_type'
  acct←acct,'i'
with_acct_type:			⍝ acct has three items
  locale←+/∧\acct[1]>tb_accounts[⍋tb_accounts[;1];1]
  acct←acct,tb_accounts[locale;4]
just_add:			⍝ acct has all four items
  ⎕es ('r'=acct[3])/'ONE MAY HAVE ONLY ONE RETAINED EARNINGS ACCOUNT.'
  ⎕es (~acct[3] ∊ 'bri')/'ACCOUNT TYPE MUST BE ONE OF b r OR i'
  ⎕es (~acct[4] ∊ 'dc')/'SIGN TYPE MUST BE ONE OF d OR c'
  tb_accounts←tb_accounts,[1]acct
  →0
∇

∇chart←tb_acct_chart
  ⍝ Function lists the chart of accounts
  chart←tb_accounts[⍋tb_accounts[;1];]
∇

∇test←tb_acct_proof acct
  ⍝ Function proves that acct (a number) exists in the chart of accounts
  ⍕(~util∆numberp acct←''⍴acct)/'test←0◊→0'
  test←∨/tb_accounts[;1]=acct
∇

∇ln←tb_acct_select amt
  ⍝ Function to append an account number on to the amount for
  ⍝ tb_doc_debit and tb_doc_credit
  ln←tb_accounts[⍋tb_accounts[;1];] tb_select amt
∇

∇wp←tb_acct_workpaper acct;dat;attr
  ⍝ Function returns a general ledger page for the acct in workpaper format
  ⎕es (~util∆numberp acct←''⍴acct)/'THE ACCOUNT MUST BE A NUMBER'
  ⎕es (~acct∊tb_accounts[;1])/(⍕acct),' NOT FOUND IN CHART OF ACCOUNTS.'
  ⎕es (0=1↑⍴dat←(acct=tb_DocLines[;3])⌿tb_DocLines)/'NO TRANSACTIONS POSTED TO ',⍕acct
  dat←tb_Documents[dat[;1];2 3 4],dat[;3 4 5]
  dat← dat,[1]'' 'Total' '' '',+⌿dat[;5 6]
  dat← 'Jrnl' 'Post Ref' 'Date' 'Acct No' 'Debit' 'Credit',[1]dat
  wp←lex∆init
  wp←wp lex∆assign 'Data' dat
  wp←wp lex∆assign (⊂'Entity'),⊂ tb_config lex∆lookup 'name'
  wp←wp lex∆assign (⊂'Title'),⊂ ' ' util∆join (⊂⍕acct),tb_accounts[tb_accounts[;1]⍳acct;2]
  wp←wp lex∆assign (⊂'Period'),tb_config lex∆lookup 'end'
  wp←wp lex∆assign (⊂'Id'),⊂'tb_gl_',⍕acct
  wp←wp lex∆assign 'Author' 'tb workspace'
  attr←(⍴dat)⍴⊂lex∆init
  attr[;4]←⊂((lex∆init)lex∆assign (⊂'format'),⊂tb_config lex∆lookup 'accountFormat')lex∆assign 'class' 'number'
  attr[;5 6]←⊂((lex∆init)lex∆assign (⊂'format'),⊂tb_config lex∆lookup 'balanceFormat')lex∆assign 'class' 'number'
  attr[1;]←⊂(lex∆init)lex∆assign 'class' 'colhead'
  attr[1↑⍴attr;5 6]←⊂((lex∆init)lex∆assign 'class' 'grand')lex∆assign (⊂'format'),⊂tb_config lex∆lookup 'balanceFormat'
  wp←wp lex∆assign 'Attributes' attr
  wp←wp lex∆assign 'Footer' ''
∇

∇tb_acct_show acct
  ⍝ Fucntion to interactively display an account and its transactions.
  ⍞←wp∆txt∆assemble tb_acct_workpaper acct
∇

∇attr←tb_attr_create dat;shape;ix;nbr_a
  ⍝ Function creates a workpaper Attribute array
  attr←(shape←⍴dat)⍴⊂lex∆init
  ix←¯1↓1↓⍳shape[1]
  →(4 9=shape[2])/tb,adj
err:
  ⎕es 'DATA IS NOT GENERATED BY tb_trialbalance'
tb:
  attr[ix;3 4]←nbr_a←⊂((lex∆init)lex∆assign 'class' 'number')lex∆assign (⊂'format'),⊂   tb_config lex∆lookup 'balanceFormat'
  attr[ix;1]←⊂((lex∆init)lex∆assign 'class' 'account') lex∆assign (⊂'format'),⊂    tb_config lex∆lookup 'accountFormat'
  attr[shape[1];3 4]←⊂((lex∆init)lex∆assign 'class' 'grand')lex∆assign (⊂'format'),⊂tb_config lex∆lookup 'balanceFormat'
  attr[1;]←⊂((lex∆init)lex∆assign 'class' 'colhead') lex∆assign 'text-align' 'center'
  →0
adj:
  attr[ix;3 4 6 7 8 9]←nbr_a←⊂((lex∆init)lex∆assign 'class' 'number')lex∆assign (⊂'format'),⊂   tb_config lex∆lookup 'balanceFormat'
  attr[ix;1]←⊂((lex∆init)lex∆assign 'class' 'account') lex∆assign (⊂'format'),⊂    tb_config lex∆lookup 'accountFormat'
  attr[shape[1];3 4 6 7 8 9]←⊂((lex∆init)lex∆assign 'class' 'grand')lex∆assign (⊂'format'),⊂tb_config lex∆lookup 'balanceFormat'
  attr[1;]←⊂((lex∆init)lex∆assign 'class' 'colhead') lex∆assign 'text-align' 'center'
  →0  
∇

∇wp←tb_chart_workpaper;dat
  ⍝ Function assemble a workpaper to display the chart of accounts.
  dat←'Acct_no' 'Title' 'Type' 'Sign',[1]tb_accounts[⍋tb_accounts[;1];]
  attr←(⍴dat)⍴⊂lex∆init
  attr[1;]←⊂(lex∆init)lex∆assign 'class' 'colhead'
  wp←lex∆init
  wp←wp lex∆assign 'Data' dat
  wp←wp lex∆assign (⊂'Entity'),⊂ tb_config lex∆lookup 'name'
  wp←wp lex∆assign (⊂'Title'),⊂ 'Chart of Accounts'
  wp←wp lex∆assign (⊂'Period'),⊂tb_config lex∆lookup 'end'
  wp←wp lex∆assign (⊂'Id'),⊂'tb_chart'
  wp←wp lex∆assign 'Author' 'tb workspace'
  wp←wp lex∆assign 'Attributes' attr
  wp←wp lex∆assign 'Footer' ''
∇
  
∇tb_chart_show
  ⍝ Function to display the chart of accounts
  ⍞←wp∆txt∆assemble tb_chart_workpaper
∇

∇doc← tb_doc_create args;rootNode;flds
  ⍝ Functions creates a document.  args is a vector of four items:
  ⍝ journal, name, date and description.
  ⎕es(4≠⍴args←,args)/'LENGTH ERROR ARGUMENT MUST BE A FOUR ITEM VECTOR OF JOURNAL, NAME, DATE, DESCRIPTION'
  doc←(⊂0, args),⊂0 5⍴0
∇

∇new←doc tb_doc_credit ln
  ⍝ Function to add or replace a line in doc with a debit.
  ⎕es (2≠⍴ln←,ln)/'LENGTH ERROR ARGUMENT MUST BE A TWO ITEM VECTOR'
  ⎕es (~tb_acct_proof ln[1])/(⍕ln[1]),' NOT IN CHART OF ACCOUNTS.'
  ln←0 0, ln[1], 0, ln[2]
  new←doc tb_doc_newLine ln
∇

∇new←doc tb_doc_debit ln
  ⍝ Function to add or replace a line in doc with a debit.
  ⎕es (2≠⍴ln←,ln)/'LENGTH ERROR ARGUMENT MUST BE A TWO ITEM VECTOR'
  ⎕es (~tb_acct_proof ln[1])/(⍕ln[1]),' NOT IN CHART OF ACCOUNTS.'
  ln←0 0, ln, 0
  new←doc tb_doc_newLine ln
∇

∇new←desc tb_doc_describe entry
  ⍝ Function to change or return the description of an entry.
  →(2=⎕nc 'desc')/change
  ⍝ Return an entry's description
  new←(⊃entry[1])[5]
  →0
change:				⍝ an entry's description
  new←entry
  new[1]←⊂(4↑⊃new[1]),⊂desc
  a →0
∇

∇doc←tb_doc_get id;head;body;bv;ix;l;dshape;jshape;nshape;err
  ⍝ Function returns a document. Argument is either the line in
  ⍝ tb_Documents or a vector of Journal, Name and Date
  →((util∆numberp id)∧0=⍴⍴id)/id_supplied
  →(3=⍴id)/lookup
  ⎕es 'DOCUMENT ID MUST BE A NUMBER OR A VECTOR OF JOURNAL, NAME AND DATE'
lookup:
  jshape←⍴⊃tb_Documents[;2]
  nshape←⍴⊃tb_Documents[;3]
  dshape←⍴⊃tb_Documents[;4]
  ⍝ Find matching journals
  bv←(⊃tb_Documents[;2])∧.=jshape[2]↑⊃id[1]
  ⍝ Find matching names
  bv←bv∧(⊃tb_Documents[;3])∧.=nshape[2]↑⊃id[2]
  ⍝ Find matching dates
  bv←bv∧(⊃tb_Documents[;4])∧.=dshape[2]↑⊃id[3]
  →(1≠+/bv)/err_search
  id←''⍴bv/tb_Documents[;1]
  →id_supplied
id_supplied:
  →(0=ix←''⍴(id=tb_Documents[;1])/⍳1↑⍴tb_Documents)/err_search
  doc←(⊂tb_Documents[ix;]),⊂(tb_DocLines[;1]=id)⌿tb_DocLines
  →0
err_search:
  err←(⍕id),' NOT IN DATABASE.'
  ⎕es err
∇

∇list←tb_doc_list
  ⍝ Function returns a list of documents
  list←(2 3 ⍴'   Date   ' ' Jrnl ' ' name ' '==========' ' ==== ' ' ==== '),[1]tb_Documents[;4 2 3]
∇

∇new←name tb_doc_name entry;head
  ⍝ Function to change or return an entry's name
  →(2=⎕nc 'name')/change
  ⍝ Return an entry's name
  new←(1⊃entry)[3]
  →0
change:			⍝ an entry's name
  head←1⊃entry
  head[3]←⊂name
  new←(⊂head),entry[2]
  →0
∇

∇new←doc tb_doc_newLine ln;docLines;ix;ct
  ⍝ Function appends a new line or replaces and old line (based on
  ⍝ account) in a document.
  docLines←⊃doc[2]
  →(0=ct←1↑⍴docLines)/add
  →(ct<ix←docLines[;3]⍳ln[3])/add
replace:
  ln[2]←ix
  docLines[ix;]←ln
  →end
add:
  ln[2]←1+''⍴⍴docLines
  docLines←docLines,[1]ln
  →end
end:
  new←doc[1],⊂docLines
∇

∇tb_doc_post doc;ix;head;body
  ⍝ Function posts a document to the database.
  ix←tb_seq_doc_id
  head←⊃doc[1]
  head[1]←ix
  body←⊃doc[2]
  body[;1]←ix
  ⎕es (≠/+⌿body[;4 5])/'THE DEBITS DO NOT EQUAL THE CREDITS'
  tb_Documents←tb_Documents,[1]head
  tb_DocLines←tb_DocLines,[1]body
∇

∇doc←old tb_doc_delLine acct;lines
  ⍝ Function to delete a line from a document.
  doc←old
  lines←⊃doc[2]
  ⎕es (~util∆numberp acct←''⍴acct)/'ACCOUNT NUMBER IS NOT A NUMBER'
  ⎕es (0≠≡acct)/'ACCOUNT NUMBER IS NOT A SINGLE NUMBER'
  ⎕es (~acct∊lines[;3])/(⍕acct),' NOT FOUND'
  doc[2]←⊂(lines[;3]≠acct)⌿lines
∇

∇wp←tb_doc_workpaper doc;aix;meta_doc;lines;attr;dat
  ⍝ Function returns a workpaper displaying a document
  meta_doc←⊃doc[1]
  lines←⊃doc[2]
  wp←lex∆init
  wp←wp lex∆assign (⊂'Entity'),⊂ tb_config lex∆lookup 'name'
  wp←wp lex∆assign (⊂'Title'),⊂ ' ' util∆join (⊂'Journal Entry'),meta_doc[2 3]
  wp←wp lex∆assign (⊂'Period'),meta_doc[4]
  wp←wp lex∆assign (⊂'Id'),meta_doc[3]
  wp←wp lex∆assign 'Author' 'tb workspace'
  dat←tb_accounts[tb_accounts[;1]⍳lines[;3];1 2],lines[;4 5]
  dat←dat,[1](⊂''),(⊂'Total'),+⌿lines[;4 5]
  wp←wp lex∆assign 'Data' dat
  wp←wp lex∆assign 'Stylesheet' wp∆defaultcss
  attr←(⍴dat)⍴⊂lex∆init
  attr[;1]←⊂((lex∆init)lex∆assign (⊂'format'),⊂tb_config lex∆lookup 'accountFormat')lex∆assign 'class' 'number'
  attr[;3 4]←⊂((lex∆init)lex∆assign (⊂'format'),⊂tb_config lex∆lookup 'balanceFormat')lex∆assign 'class' 'number'
  wp←wp lex∆assign 'Attributes' attr
  wp←wp lex∆assign (⊂'Footer'), meta_doc[5]
∇


∇tb_doc_show doc
  ⍝ Function to display an document in general journal form
  ⍞←wp∆txt∆assemble tb_doc_workpaper doc
∇

∇tb_journal_show jrnl;ix
  ⍝ Function to display a journal in general journal form.  jrnl is the key to the tb_journal lexicon
  ix←((⊃tb_Documents[;2])∧.=2↑,jrnl)/⍳1↑⍴tb_Documents
  ⍞←⎕av[11],¨ wp∆txt∆assemble ¨ tb_doc_workpaper ¨ tb_doc_get ¨ ix
∇

∇flds tb_import_delimited fname;delim;b1;b2;mag;name;ye;re;bv;rdat
  ⍝ Function to import a delimited trial balance. Left argument flds
  ⍝ is a vector columnar indicies to the trial balance array in fname:
  ⍝ AcctNo AcctTile Dr Cr
  ⍎(2≠⎕nc'flds')/'flds←⍳4'
  ⍝delim←flds[1]
  ⍝flds←1↓flds
  ⎕es (2≠⎕nc'tb_config')/'You must initiate the data base before importing data'
  b1←~∧/(tb_config lex∆haskey 'name'),(tb_config lex∆haskey 'end'),tb_config lex∆haskey 'retainedEarnings'
  ⎕es b1/'company name, end, and retainedEarnings must all be defined in tb_config before importing data'
  name←tb_config lex∆lookup 'name'
  ye←tb_config lex∆lookup 'end'
  re←tb_config lex∆lookup 'retainedEarnings'
  dat←import∆table import∆file fname
  ⍝ Remove descriptive material
  ⍝rdat←⍴dat
  ⍝bv←2≤+/rdat⍴(util∆numberp¨dat)/dat←,dat
  ⍝dat←bv⌿rdat⍴dat
  ⍝ Build tb_accounts
  tb_accounts←(dat[;flds[1 2]],'b'),'d'		⍝ Account type and sign type
  ⍝ Mark retained earnings
  tb_accounts[(tb_accounts[;1]=re)/⍳1↑⍴tb_accounts;3]←'r'
  ⍝ Mark income accounts
  tb_accounts[(tb_accounts[;1]>re)/⍳1↑⍴tb_accounts;3]←'i'
  ⍝ Build document
  tb_Documents←1 5 ⍴ (tb_seq_doc_id),(⊂'gj'),(⊂'BegBal'),(⊂ye),⊂'Begining balances for ',name
  ⍝ Build doc lines
  →(3=⍴flds)/threeCols
fourCols:
  ⍝             Doc_id        line no   AcctNo      Dr Cr
  tb_DocLines←tb_Documents[1;1],(⍳1↑⍴dat),dat[;flds[1 3  4]]
  →rest
threeCols:
  b1←dat[;flds[3]]>0
  b2←dat[;flds[3]]<0
  ⍝              Doc_id         line No    AcctNo        Dr                  Cr
  tb_DocLines←tb_Documents[1;1],(⍳1↑⍴dat),dat[;flds[1]],(b1\b1/dat[;flds[3]]),[1.1]b2\b2/-dat[;flds[3]]
  →rest
rest:
  mag←⌈10⍟⌈/tb_DocLines[;3]	⍝ For the width of a column of acct numbers
  tb_config←tb_config lex∆assign (⊂'accountFormat'),⊂mag⍴'0'
∇

∇tb_init meta
  ⍝ Function to set up the tb database. meta is a lexicon of company
  ⍝ (name), end (year end date) and retainedEarnings (account number).
  tb_DictAccounts←'Acct No' 'Title' 'Acct Type' 'Sign Type'
  tb_DictDocuments← 'doc_id' 'Journal' 'Name' 'Date' 'Description'
  tb_Documents←(0,⍴tb_DictDocuments)⍴0
  tb_DictDocLines←'doc_id' 'line no' 'Acct No' 'Debit' 'Credit'
  tb_DocLines←(0,⍴tb_DictDocLines)⍴0
  tb_seq_doc_id_next←1
  →(lex∆is meta)/lexArg
vectorArg:
  ⎕es (1≠⍴⍴meta)/'ARGUMENT NOT A THREE ELEMENT VECTOR'
  ⎕es (3≠⍴meta)/'ARGUMENT NOT A THREE ELEMENT VECTOR'
  tb_config←(((lex∆init)lex∆assign 'name' meta[1])lex∆assign 'end' meta[2])lex∆assign 'retainedEarnings' meta[3]
  →completeConfig
lexArg:
  tb_config←meta
  →completeConfig
completeConfig:
  tb_config←tb_config lex∆assign 'accountFormat' '0000'
  tb_config←tb_config lex∆assign 'balanceFormat' '55,555,550.00'
  tb_config←tb_config lex∆assign 'BeginingJE' 'BEGBAL'
  tb_accounts←(1,⍴tb_DictAccounts)⍴ (tb_config lex∆lookup 'retainedEarnings'), (⊂'Retained earnings'), 'r' 'c'
  tb_initStyle
∇

∇ans←chart tb_select args;shape;mag;keys;max;min
  ⍝ Function prepends the item selected from chart to the list of args
  →(util∆numberp keys←chart[;1])/mm
  keys←⊃⍎¨chart[;1]
mm:
  max←⌈/keys
  min←⌊/keys
  shape←1↑⍴chart
  mag←1 + ⌊ 10⍟shape
  ((('[',mag⍴'0'),']')⍕(shape,1)⍴⍳shape),chart
pt:
  ans←prompt 'Select [item] '
  →(util∆numberp ans)/keyword
  ans←(~∧\' '=ans)/ans
  ⍎(~util∆numberis ans)/'''Please enter a number less than '',⍕shape◊→pt'
  ans←⍎ans
  →((ans≤max)∧ans≥min)/key
index:
  ⍎(shape < ans)/'''Please enter a number less than '',⍕shape◊→pt'
  ans←chart[ans;1],args
  →0
key:
  ans←(ans=keys)⌿chart[;1]
  ⍎(1≠⍴ans)/'''Invalid key''◊→pt'
  ans←ans,args
  →0
keyword:
  →(0 ≠ ans)/pt
  →0
∇

∇next←tb_seq_doc_id
  ⍝ Function returns the next item in the doc_id sequence
  next←tb_seq_doc_id_next
  tb_seq_doc_id_next ← 1 + tb_seq_doc_id_next
∇

∇bg←tb_begin_init
  ⍝ Function to initiate a begining balance trial balance. Begining
  ⍝ balances dictionary is acct_no title debit credit.
  bg←0 4⍴0
∇

∇bg←old tb_begin_newline acct;ix
  ⍝ Function to add an account to a begining balance trial balance
  bg←old
  →((1↑⍴bg)<ix←bg[;1]⍳acct[1])/add
replace:
  bg[ix;]←acct
  →0
add:
  bg←bg,[1]acct
  →0
∇

∇b←tb_begin_editCheck acct
  ⍝ Function to test data before adding/inserting into a begining
  ⍝ trial balance
  →(~b←4=⍴acct←,acct)/0		⍝ Must be four items
  →(~b←∧/util∆numberp ¨ acct[1 3 4])/0
  b←util∆stringp ⊃acct[2]
∇

∇bg←old tb_begin_debit acct
  ⍝ Function to add or insert an account with a debit balance to a
  ⍝ begining trial balance
  acct←acct,0
  ⎕es (~tb_begin_editCheck acct)/'A Trial balance debit must be an account no, title, and debit'
  bg←old tb_begin_newline acct
∇

∇bg←old tb_begin_credit acct
  ⍝ Function to add or insert an account with a credit balance to a
  ⍝ begining trial balance.
  acct←acct[1 2],0,acct[3]
  ⎕es(~tb_begin_editCheck acct)/'A Trial balance credit must be an account no, title, and credit'
  bg←old tb_begin_newline acct
∇

∇tb_begin_post bg;ix;je;ax;bv;re;if
  ⍝ Function posts a begining trial balance to the workspace.
  ⎕es (~=/+⌿bg[;3 4])/'DEBITS DO NOT EQUAL CREDITS'
  ix←tb_accounts[;1]⍳bg[;1]
  ax←(bv←ix>1↑⍴tb_accounts)/⍳⍴ix
  re←tb_config lex∆lookup 'retainedEarnings'
  if←'bi'[⎕io+bg[ax;1]>re]
  tb_accounts←tb_accounts,[1](bg[ax;1 2],if),'d'
  je←tb_doc_create 'gj',((⊂tb_config) lex∆lookup ¨ 'BeginingJE' 'end'),''
  je[2]←⊂je[1],(⍳1↑⍴bg),bg[;1 3 4]
  tb_doc_post je
∇

∇sched←tb_sched_init proto
  ⍝ Function returns a new schedule. Proto is a vector of id and
  ⍝ title.
  ⎕es (~util∆stringp 1⊃proto)/'SCHEDULE ID MUST BE A CHARACTER STRING.'
  ⎕ex (~util∆stringp 2⊃proto)/'SCHEDULE TITLE MUST BE A CHJARACTER STRING.'
  sched←lex∆init
  sched←sched lex∆assign (⊂'ID' ),proto[1]
  sched←sched lex∆assign (⊂'title'),proto[2]
  sched←sched lex∆assign (⊂'lines'),⊂0 6⍴0
  sched←sched lex∆assign 'fns' 'tb_sched_workpaper'
  sched←sched lex∆assign 'title' 'A Schedule'
∇

∇sched←sched tb_sched_addLine line;lines
  ⍝ Function adds or replaces a line
  lines←sched lex∆lookup 'lines'
  lines←lines,[1]line
  sched←sched lex∆assign 'lines' lines
∇

∇rpt←tb_sched_showLines sched;lines
  ⍝ Function returns a report of the name, line number and title of
  ⍝ each line on the schedule.
  lines←sched lex∆lookup 'lines'
  rpt←lines[⍋lines[;2];1 2 3]
∇

∇wp←tb_sched_workpaper sched;dat;attr;lines
  ⍝ Function prepares a detail schedule
  wp←lex∆init
  wp←wp lex∆assign (⊂'Entity'),⊂ tb_config lex∆lookup 'name'
  wp←wp lex∆assign (⊂'Title'),⊂ sched lex∆lookup 'title'
  wp←wp lex∆assign (⊂'Period'),⊂ tb_config lex∆lookup 'end'
  wp←wp lex∆assign (⊂'Id'),⊂'A Schedule'
  wp←wp lex∆assign 'Author' 'tb workspace'
  lines←sched lex∆lookup 'lines'
  dat←tb_sched_workpaper1 lines[⍋lines[;2];]
  wp←wp lex∆assign 'Data' dat
  attr←(⍴dat)⍴⊂lex∆init
  attr[;2]←⊂((lex∆init)lex∆assign 'class' 'number')lex∆assign (⊂'format'),⊂tb_config lex∆lookup 'accountFormat'
  attr[;4]←⊂((lex∆init) lex∆assign 'class' 'number')lex∆assign (⊂'format'),⊂tb_config lex∆lookup 'balanceFormat'
  ix←(bv←(⊃dat[;3])[;⍳5]∧.='Total')/⍳1↑⍴dat
  attr[ix;2]←⊂lex∆init
  attr[ix;4]←⊂((lex∆init)lex∆assign 'class' 'total')lex∆assign (⊂'format'),⊂tb_config lex∆lookup 'balanceFormat'
  ix←(¯1⌽bv)/⍳1↑⍴dat
  attr[ix;2 4]←⊂lex∆init
  wp←wp lex∆assign 'Attributes' attr
∇

∇dat←tb_sched_workpaper1 lines;ix;accts;grp;ln1;lnTotal;bal;subs;space
  ⍝ Function returns a data array with lines added for each account as
  ⍝ well as a Heading decorator and total line. Colmuns returned are
  ⍝ GRP code, Acct No, Acct Title, Balance.
  →(0<1↑⍴lines)/compile
  dat←0 5⍴0
  →0
compile:
  space←⊂' '
  dat←tb_sched_workpaper1 1 0↓lines
  grp←lines[1;]
  →(0=⍴⍴accts←⊃grp[5])/lineTotals
acctTotals:
  accts←accts[⍋accts]
  ix←tb_accounts[;1]⍳accts
  ln1←grp[1],space,grp[3],2⍴space
  lnTotal←+⌿bal←tb_tb_balances[ix;]
  lnTotal←(2⍴space),(⊂'Total ',⊃grp[1]),lnTotal
  dat←ln1,[1](grp[1],tb_accounts[ix;1 2],bal),[1]lnTotal,[1]dat
  →0
lineTotals:
  dat←(grp[⍳3],2⍴space),[1]dat
  →0
∇

∇dat←raw tb_sched_workpaper2 lines;ix;curLn;lnTotal
  ⍝ Function calculates total lines from the raw report (ie just
  ⍝ totals for accounts.
  dat←raw
  →(0≥1↑⍴lines)/0		⍝ End recusrion
  dat←tb_sched_workpaper2 1 0 ↓ lines
  curLn←lines[1;]
  lnTotal←+⌿raw
  
∇

⍝ There are three type of lines on a schedule:
⍝  1) Decoration has no amounts
⍝  2) Account has balances from a trial balance.
⍝  3) Totals has balances that are totals of either accounts or other
⍝  total lines. 

∇new←sched tb_line_init proto;subs;lines;line
  ⍝ Function to initiate a schedule line.  Proto is a vector of name,
  ⍝ line_no, title balance_flag ('d' or 'c') and either a list of
  ⍝ acct_no or line_no.
  line←4↑proto
  subs←4↓proto
  ⎕es (~util∆stringp 1⊃proto)/'LINE NAME MUST BE A CHARACTER STRING'
  ⎕es (~util∆numberp 2⊃proto)/'LINE NUMBER MUST BE AN INTEGER'
  ⎕es (~util∆stringp 3⊃proto)/'TITLE MUST BE A CHARACTER STRING'
  ⎕es (~(4⊃proto)∊'cd')/'BALANCE FLAG MUST BE EITHER ''c'' or ''d'''
  →(∧/(1↑⍴tb_accounts)≥tb_accounts[;1]⍳subs)/add_lines
  lines←sched lex∆lookup 'lines'
  →(∧/(1↑⍴lines)≥lines[;2]⍳subs)/add_subs
  ⎕es (⍕subs),' NOT FOUND IN CHART OF ACCOUNTS OR SCHEDULE'
add_lines: new←sched tb_sched_addLine line,(⊂subs),⊂' '
  →0
add_subs: new←sched tb_sched_addLine line,(⊂' '),⊂subs
  →0
 ∇  

∇dat←tb_tb_balances;shape
  ⍝ trial balance helper function to calculate closing balances
  shape←(1↑⍴tb_accounts),1↑⍴tb_DocLines
  dat←-/tb_DocLines[;4 5]
  dat←+/(shape⍴dat)×tb_accounts[;1]∘.=tb_DocLines[;3]
  dat←¯2 util∆round (dat×dat>0),[1.1]-dat×dat<0
∇

∇tb_tb_show                          
  ⍝ Function display a trial balance
  ⍞←wp∆txt∆assemble tb_trialbalance 
∇

∇tb_tb_adj_show                                   
  ⍝ Function to display an adjusted trial balance
  ⍞←wp∆txt∆assemble tb_trialbalance_adj          
∇

∇tb←tb_trialbalance;dat
  ⍝ Function returns a trial balance
  dat←tb_accounts[;1 2],tb_tb_balances
  dat←dat[⍋dat[;1];]
  dat←dat,[1]'' 'Total',+⌿dat[;3 4]
  dat← 'AcctNo' 'Title' 'Dr' 'Cr' ,[1] dat
  ⍝ Create a workpaper
  tb←lex∆init
  tb←tb lex∆assign 'Data' dat
  tb←tb lex∆assign (⊂'Id'),⊂'TB_',tb_config lex∆lookup 'end'
  tb←tb lex∆assign (⊂'Entity'),⊂ tb_config lex∆lookup 'name'
  tb←tb lex∆assign 'Title' 'Trial Balance'
  tb←tb lex∆assign (⊂'Period' ),⊂ tb_config lex∆lookup 'end'
  tb←tb lex∆assign 'Author' 'tb workspace'
  tb←tb lex∆assign 'Function' ''
  tb←tb lex∆assign 'Stylesheet' wp∆defaultcss
  tb←tb lex∆assign (⊂'Attributes'),⊂ tb_attr_create dat
∇

∇tb←tb_trialbalance_adj;dat;b1;b2;t1;t2;t3
  ⍝ Function returns a workpaper in the form of an adjusted trial balance.
  ⍝ Assemble the trial balance. dat dictionary is
  ⍝ 1) Account number
  ⍝ 2) Title
  ⍝ 3) Begining Debits
  ⍝ 4) Begining Credits
  ⍝ 5) Posting reference (name)
  ⍝ 6) Adjusting Dr
  ⍝ 7) Adjusting Cr
  dat←(tb_accounts[⍋tb_accounts[;1];1 2],0),0
  t1←(b1←tb_DocLines[;1]=1)⌿tb_DocLines[;3 4 5]
  t2←(~b1)⌿tb_DocLines
  t2←t2,tb_Documents[tb_Documents[;1]⍳t2[;1];3]
  dat[dat[;1]⍳t1[;1];3 4]←t1[;2 3]
  t2←t2[⍋t2[;3];]
  t3←(b2←1,(1↓t2[;3])≠¯1↓t2[;3])⌿t2
  t2←(~b2)⌿t2
  dat←((dat,⊂''),0),0
  dat[dat[;1]⍳t3[;3];1 5 6 7]←t3[;3 6 4 5]
  dat←dat,[1]t2[;3],((⌽3,1↑⍴t2)⍴'' 0 0),t2[;6 4 5]
  dat←dat[⍋dat[;1];]
  b1←1,(¯1↓dat[;1])≠1↓dat[;1]
  dat←dat,b1⍀tb_tb_balances[⍋tb_accounts[;1];]
  dat←dat,[1]('' 'Total'),( +⌿dat[;3 4]), (⊂''), +⌿dat[;6 7 8 9]
  dat←'AcctNo' 'Acct Title' 'Begin Dr' 'Begin Cr' 'Ref' 'Dr' 'Cr' 'Adj Dr' 'Adj Cr',[1] dat
  tb←(lex∆init) lex∆assign 'Data' dat
  tb←tb lex∆assign (⊂'Id'),⊂'TB_',tb_config lex∆lookup 'end'
  tb←tb lex∆assign (⊂'Entity'),⊂ tb_config lex∆lookup 'name'
  tb←tb lex∆assign 'Title' 'Trial Balance'
  tb←tb lex∆assign (⊂'Period' ),⊂ tb_config lex∆lookup 'end'
  tb←tb lex∆assign 'Author' 'tb workspace'
  tb←tb lex∆assign 'Function' ''
  tb←tb lex∆assign 'Stylesheet' wp∆defaultcss
  tb←tb lex∆assign (⊂'Attributes'),⊂ tb_attr_create dat
∇

∇tb_initStyle;css
  ⍝ Function to initiate a trial balance specific style sheet
  css←lex∆init
  css←css lex∆assign (⊂'body'),⊂(((lex∆init)lex∆assign 'font' '12pt sans-serif')lex∆assign 'text-align' 'left')lex∆assign 'border' 'none 2pt black'

  css←css lex∆assign (⊂'.colhead'),⊂((lex∆init)lex∆assign 'text-align' 'center')lex∆assign 'font-size' 'small'

  css←css lex∆assign (⊂'.number'),⊂((lex∆init)lex∆assign 'text-align' 'right')lex∆assign 'width' '9em'

  css←css lex∆assign (⊂'.account'),⊂(lex∆init)lex∆assign 'test-align' 'right'

  css←css lex∆assign (⊂'.page-head'),⊂((lex∆init)lex∆assign 'font-size' 'large') lex∆assign 'font-weight' 'bold'
  
  css←css lex∆assign (⊂'.initial-block'),⊂((lex∆init)lex∆assign 'font-size' 'small')lex∆assign 'border' 'solid 1pt black'

  css←css lex∆assign(⊂'.grand'),⊂(((lex∆init)lex∆assign 'border-top' 'solid 1pt black')lex∆assign 'border-bottom' 'double 3pt black')lex∆assign 'text-align' 'right'
  css←css lex∆assign(⊂'.total'),⊂((lex∆init)lex∆assign 'border-top' 'solid 1pt black')lex∆assign 'text-align' 'right'
  wp∆defaultcss←css
∇

tb_data_vars←'tb_DictAccounts' 'tb_accounts' 'tb_DictDocuments' 'tb_Documents' 'tb_DictDocLines' 'tb_DocLines' 'tb_seq_doc_id' 'tb_config' 'wp∆defaultcss'
